package java.util.concurrent.atomic;

import java.util.function.IntBinaryOperator;
import java.util.function.IntUnaryOperator;
import sun.misc.Unsafe;
import java.io.Serializable;

public class AtomicIntegerArray implements Serializable
{
    private static final long serialVersionUID = 2862133569453604235L;
    private static final Unsafe unsafe;
    private static final int base;
    private static final int shift;
    private final int[] array;
    
    private long checkedByteOffset(int i) {
        if (i < 0 || i >= this.array.length) {
            throw new IndexOutOfBoundsException("index " + i);
        }
        return byteOffset(i);
    }
    
    private static long byteOffset(int n) {
        return ((long)n << AtomicIntegerArray.shift) + (long)AtomicIntegerArray.base;
    }
    
    public AtomicIntegerArray(int n) {
        this.array = new int[n];
    }
    
    public AtomicIntegerArray(int[] array) {
        this.array = (int[])array.clone();
    }
    
    public final int length() {
        return this.array.length;
    }
    
    public final int get(int n) {
        return this.getRaw(this.checkedByteOffset(n));
    }
    
    private int getRaw(long offset) {
        return AtomicIntegerArray.unsafe.getIntVolatile(this.array, offset);
    }
    
    public final void set(int n, int x) {
        AtomicIntegerArray.unsafe.putIntVolatile(this.array, this.checkedByteOffset(n), x);
    }
    
    public final void lazySet(int n, int x) {
        AtomicIntegerArray.unsafe.putOrderedInt(this.array, this.checkedByteOffset(n), x);
    }
    
    public final int getAndSet(int n, int newValue) {
        return AtomicIntegerArray.unsafe.getAndSetInt(this.array, this.checkedByteOffset(n), newValue);
    }
    
    public final boolean compareAndSet(int n, int n2, int n3) {
        return this.compareAndSetRaw(this.checkedByteOffset(n), n2, n3);
    }
    
    private boolean compareAndSetRaw(long offset, int expected, int x) {
        return AtomicIntegerArray.unsafe.compareAndSwapInt(this.array, offset, expected, x);
    }
    
    public final boolean weakCompareAndSet(int n, int n2, int n3) {
        return this.compareAndSet(n, n2, n3);
    }
    
    public final int getAndIncrement(int n) {
        return this.getAndAdd(n, 1);
    }
    
    public final int getAndDecrement(int n) {
        return this.getAndAdd(n, -1);
    }
    
    public final int getAndAdd(int n, int delta) {
        return AtomicIntegerArray.unsafe.getAndAddInt(this.array, this.checkedByteOffset(n), delta);
    }
    
    public final int incrementAndGet(int n) {
        return this.getAndAdd(n, 1) + 1;
    }
    
    public final int decrementAndGet(int n) {
        return this.getAndAdd(n, -1) - 1;
    }
    
    public final int addAndGet(int n, int n2) {
        return this.getAndAdd(n, n2) + n2;
    }
    
    public final int getAndUpdate(int n, IntUnaryOperator intUnaryOperator) {
        long checkedByteOffset = this.checkedByteOffset(n);
        int raw;
        do {
            raw = this.getRaw(checkedByteOffset);
        } while (!this.compareAndSetRaw(checkedByteOffset, raw, intUnaryOperator.applyAsInt(raw)));
        return raw;
    }
    
    public final int updateAndGet(int n, IntUnaryOperator intUnaryOperator) {
        long checkedByteOffset = this.checkedByteOffset(n);
        int raw;
        int applyAsInt;
        do {
            raw = this.getRaw(checkedByteOffset);
            applyAsInt = intUnaryOperator.applyAsInt(raw);
        } while (!this.compareAndSetRaw(checkedByteOffset, raw, applyAsInt));
        return applyAsInt;
    }
    
    public final int getAndAccumulate(int n, int n2, IntBinaryOperator intBinaryOperator) {
        long checkedByteOffset = this.checkedByteOffset(n);
        int raw;
        do {
            raw = this.getRaw(checkedByteOffset);
        } while (!this.compareAndSetRaw(checkedByteOffset, raw, intBinaryOperator.applyAsInt(raw, n2)));
        return raw;
    }
    
    public final int accumulateAndGet(int n, int n2, IntBinaryOperator intBinaryOperator) {
        long checkedByteOffset = this.checkedByteOffset(n);
        int raw;
        int applyAsInt;
        do {
            raw = this.getRaw(checkedByteOffset);
            applyAsInt = intBinaryOperator.applyAsInt(raw, n2);
        } while (!this.compareAndSetRaw(checkedByteOffset, raw, applyAsInt));
        return applyAsInt;
    }
    
    @Override
    public String toString() {
        int n = this.array.length - 1;
        if (n == -1) {
            return "[]";
        }
        StringBuilder sb = new StringBuilder();
        sb.append('[');
        int n2 = 0;
        while (true) {
            sb.append(this.getRaw(byteOffset(n2)));
            if (n2 == n) {
                break;
            }
            sb.append(',').append(' ');
            ++n2;
        }
        return sb.append(']').toString();
    }
    
    static {
        unsafe = Unsafe.getUnsafe();
        base = AtomicIntegerArray.unsafe.arrayBaseOffset(int[].class);
        int arrayIndexScale = AtomicIntegerArray.unsafe.arrayIndexScale(int[].class);
        if ((arrayIndexScale & arrayIndexScale - 1) != 0x0) {
            throw new Error("data type scale not a power of two");
        }
        shift = 31 - Integer.numberOfLeadingZeros(arrayIndexScale);
        return;
    }
}
